@@ -12,6 +12,7 @@ __all__ = [
             'enumerate_layers',
             'extract_features',
             'get_cls_mapping_imgnet',
+            'get_imagenet_classes',
             'get_digits',
             'get_model',
             'get_shape',
@@ -43,6 +44,7 @@ import os
 import random
 import re
 import scipy
+import scipy.io
 import torch
 import itertools
 
@@ -78,6 +80,9 @@ def load_dl(
              transforms=None,
              ) -> Iterator:
     print(f'\n...Loading dataset into memory.')
+    if not os.path.exists(out_path):
+        os.makedirs(out_path)
+        print(f'...Creating output directory.')
     dataset = ImageDataset(
                             root=root,
                             out_path=out_path,
@@ -91,15 +96,15 @@ def load_dl(
     dl = DataLoader(dataset, batch_size=batch_size, shuffle=False)
     return dl
 
-def load_model(model_name:str, pretrained:bool, device:torch.device, model_path:str=None) -> Tuple:
+def load_model(model_name:str, pretrained:bool, device:str, model_path:str=None) -> Tuple[Any, Any]:
     """load a pretrained *torchvision* or CLIP model into memory"""
     if re.search(r'^clip', model_name):
-        assert isinstance(device, str), '\nFor CLIP models, device must be passed as a str instance.\n'
         if re.search(r'ViT$', model_name):
-            model, transforms = clip.load("ViT-B/32", device=device, model_path=model_path, jit=False)
+            model, transforms = clip.load("ViT-B/32", device=device, model_path=model_path, pretrained=pretrained, jit=False)
         else:
-            model, transforms = clip.load("RN50", device=device, model_path=model_path, jit=False)
+            model, transforms = clip.load("RN50", device=device, model_path=model_path, pretrained=pretrained, jit=False)
     else:
+        device = torch.device(device)
         if re.search(r'^cornet', model_name):
             try:
                 model = getattr(cornet, f'cornet_{model_name[-1]}')
@@ -139,45 +144,45 @@ def get_module_names(model, module:str) -> list:
 
 def extract_features_across_models_and_datasets(
                                                 out_path:str,
-                                                model_names:list,
-                                                img_paths:list,
-                                                module_names:list,
+                                                model_names:List[str],
+                                                img_paths:List[str],
+                                                module_names:List[str],
+                                                clip:List[bool],
                                                 pretrained:bool,
                                                 batch_size:int,
                                                 flatten_acts:bool,
                                                 f_format:str='.txt',
-                                                clip:bool=False,
 ) -> None:
     device = 'cuda' if torch.cuda.is_available() else 'cpu'
-    device = device if clip else torch.device(device)
     for i, model_name in enumerate(model_names):
         model, transforms = load_model(model_name, pretrained=pretrained, model_path=None, device=device)
         for img_path in img_paths:
-            dl = load_dl(img_path, batch_size=batch_size, transforms=transforms)
-            features, _ = extract_features(model, dl, module_names[i], batch_size=batch_size, flatten_acts=flatten_acts, device=device, clip=clip)
-            save_features(features, os.path.join(out_path, img_path, model_name, module_names[i], 'features'), f_format)
+            PATH = os.path.join(out_path, img_path, model_name, module_names[i], 'features')
+            dl = load_dl(img_path, out_path=PATH, batch_size=batch_size, transforms=transforms)
+            features, _ = extract_features(model, dl, module_names[i], batch_size=batch_size, flatten_acts=flatten_acts, device=device, clip=clip[i])
+            save_features(features, out_path, f_format)
 
 def extract_features_across_models_datasets_and_modules(
                                                         out_path:str,
-                                                        model_names:list,
-                                                        img_paths:list,
-                                                        module_names:list,
+                                                        model_names:List[str],
+                                                        img_paths:List[str],
+                                                        module_names:List[str],
+                                                        clip:List[str],
                                                         pretrained:bool,
                                                         batch_size:int,
                                                         flatten_acts:bool,
                                                         f_format:str='.txt',
-                                                        clip:bool=False,
 ) -> None:
     device = 'cuda' if torch.cuda.is_available() else 'cpu'
-    device = device if clip else torch.device(device)
     for i, model_name in enumerate(model_names):
         model, transforms = load_model(model_name, pretrained=pretrained, model_path=None, device=device)
         modules = get_module_names(model, module_names[i])
         for img_path in img_paths:
-            dl = load_dl(img_path, batch_size=batch_size, transforms=transforms)
             for module_name in modules:
-                features, _ = extract_features(model, dl, module_name, batch_size=batch_size, flatten_acts=flatten_acts, device=device, clip=clip)
-                save_features(features, os.path.join(out_path, img_path, model_name, module_name, 'features'), f_format)
+                PATH = os.path.join(out_path, img_path, model_name, module_name, 'features')
+                dl = load_dl(img_path, out_path=PATH, batch_size=batch_size, transforms=transforms)
+                features, _ = extract_features(model, dl, module_name, batch_size=batch_size, flatten_acts=flatten_acts, device=device, clip=clip[i])
+                save_features(features, PATH, f_format)
 
 def get_activation(name):
     """store hidden unit activations at each layer of model"""
@@ -243,8 +248,9 @@ def extract_features(
                     module_name:str,
                     batch_size:int,
                     flatten_acts:bool,
-                    device:torch.device,
+                    device:str,
                     clip:bool=False,
+                    return_predictions:bool=False,
                     feature_extractor=None,
 ) -> Tuple[np.ndarray, np.ndarray]:
     """extract hidden unit activations (at specified layer) for every image in the database"""
@@ -255,12 +261,19 @@ def extract_features(
             feature_extractor = nn.Conv2d
         else:
             feature_extractor = nn.MaxPool2d
+
+    device = torch.device(device)
     #initialise dictionary to store hidden unit activations on the fly
     global activations
     activations = {}
     #register forward hook to store activations
     model = register_hook(model)
     features, targets = [], []
+
+    if return_predictions:
+        assert not clip, '\nCannot extract activations for CLIP and return class predictions simultaneously. This feature will be implemented in a future version.\n'
+        predictions = []
+
     with torch.no_grad():
         for i, batch in enumerate(data_loader):
             batch = (t.to(device) for t in batch)
@@ -271,6 +284,10 @@ def extract_features(
                     assert torch.unique(activations[module_name] == img_features).item(), '\nImage features should represent activations in last encoder layer.\n'
             else:
                 out = model(X)
+                if return_predictions:
+                    probas = F.softmax(out, dim=1)
+                    predictions.extend(torch.argmax(probas, dim=1).numpy())
+
             if re.search(r'ensemble$', module_name):
                 layers = enumerate_layers(model, feature_extractor)
                 act = ensemble_featmaps(activations, layers, 'max')
@@ -290,20 +307,39 @@ def extract_features(
     #stack each mini-batch of hidden activations to obtain an N x F matrix, and flatten targets to yield vector
     features = np.vstack(features)
     targets = np.asarray(targets).ravel()
-    assert len(features) == len(targets)
-    print(f'\nFeatures successfully extracted for all {len(features)} images in the database\n')
+    print(f'...Features successfully extracted for all {len(features)} images in the database.')
+    if return_predictions:
+        predictions = np.asarray(predictions).ravel()
+        assert len(features) == len(targets) == len(predictions), '\nFeatures, targets, and predictions must correspond to the same number of images.\n'
+        return features, targets, predictions
+    assert len(features) == len(targets), '\nFeatures and targets must correspond to the same number of images.\n'
     return features, targets
 
 ########################################################################################################
 ################ HELPER FUNCTIONS FOR SAVING, MERGING AND SLICING FEATURE MATRICES #####################
 ########################################################################################################
 
+def rm_suffix(img:str) -> str:
+    return re.sub(r'(.eps|.jpg|.jpeg|.png|.PNG|.tif|.tiff)$', '', img)
+
 def store_features(PATH:str, features:np.ndarray, file_format:str) -> None:
+    if not os.path.exists(PATH):
+        print(f'...Output directory did not exist. Creating directories to save features.')
+        os.makedirs(PATH)
     if re.search(r'npy', file_format):
         with open(pjoin(PATH, 'features.npy'), 'wb') as f:
             np.save(f, features)
+    elif re.search(r'mat', file_format):
+        try:
+            with open(pjoin(PATH, 'file_names.txt'), 'r') as f:
+                file_names = [rm_suffix(l.strip()) for l in f]
+            features = {file_name: feature for file_name, feature in zip(file_names, features)}
+            scipy.io.savemat(pjoin(PATH, 'features.mat'), features)
+        except FileNotFoundError:
+            scipy.io.savemat(pjoin(PATH, 'features.mat'), {'features': features})
     else:
         np.savetxt(pjoin(PATH, 'features.txt'), features)
+    print(f'...Features successfully saved to disk.\n')
 
 def tensor2slices(PATH:str, file_name:str, features:np.ndarray) -> None:
     with open(pjoin(PATH, file_name), 'w') as outfile:
@@ -343,12 +379,25 @@ def slices2tensor(PATH:str, file:str) -> np.ndarray:
     return tensor
 
 def split_features(PATH:str, features:np.ndarray, file_format:str, n_splits:int) -> None:
+    if re.search(r'mat', file_format):
+        try:
+            with open(pjoin(PATH, 'file_names.txt'), 'r') as f:
+                file_names = [rm_suffix(l.strip()) for l in f]
+        except FileNotFoundError:
+            file_names = None
     splits = np.linspace(0, len(features), n_splits, dtype=int)
     for i in range(1, len(splits)):
         feature_split = features[splits[i-1]:splits[i]]
         if re.search(r'npy', file_format):
             with open(pjoin(PATH, f'features_{i:02d}.npy'), 'wb') as f:
                 np.save(f, feature_split)
+        elif re.search(r'mat', file_format):
+            if file_names:
+                file_name_split = file_names[splits[i-1]:splits[i]]
+                features = {file_name_split[i]: feature for i, feature in enumerate(feature_split)}
+                scipy.io.savemat(pjoin(PATH, f'features_{i:02d}.mat'), features)
+            else:
+                scipy.io.savemat(pjoin(PATH, f'features_{i:02d}.mat'), {'features': features})
         else:
             np.savetxt(pjoin(PATH, f'features_{i:02d}.txt'), feature_split)
 
@@ -358,6 +407,8 @@ def merge_features(PATH:str, file_format:str) -> np.ndarray:
     feature_splits = feature_splits[np.argsort(enumerations)]
     if file_format == '.txt':
         features = np.vstack([np.loadtxt(pjoin(PATH, feature)) for feature in feature_splits])
+    elif file_format == '.mat':
+        features = np.vstack([scipy.io.loadmat(pjoin(PATH, feature))['features'] for feature in feature_splits])
     elif file_format == '.npy':
         features = []
         for feature in feature_splits:
@@ -365,23 +416,23 @@ def merge_features(PATH:str, file_format:str) -> np.ndarray:
                 features.append(np.load(f))
         features = np.vstack(features)
     else:
-        raise Exception('Can only process .npy or .txt files')
+        raise Exception('\nCan only process .npy, .mat, or .txt files.\n')
     return features
 
-def parse_imagenet_synsets(file_name:str, folder:str='./data/'):
+def parse_imagenet_synsets(PATH:str):
     def parse_str(str):
         return re.sub(r'[^a-zA-Z]', '', str).rstrip('n').lower()
     imagenet_synsets = []
-    with open(pjoin(folder, file_name), 'r') as f:
+    with open(PATH, 'r') as f:
         for i, l in enumerate(f):
             l = l.split('_')
             cls = '_'.join(list(map(parse_str, l)))
             imagenet_synsets.append(cls)
     return imagenet_synsets
 
-def parse_imagenet_classes(file_name:str, folder:str='./data/'):
+def parse_imagenet_classes(PATH:str):
     imagenet_classes = []
-    with open(pjoin(folder, file_name), 'r') as f:
+    with open(PATH, 'r') as f:
         for i, l in enumerate(f):
             l = l.strip().split()
             cls = '_'.join(l[1:]).rstrip(',').strip("'").lower()
@@ -393,19 +444,24 @@ def parse_imagenet_classes(file_name:str, folder:str='./data/'):
 def get_class_intersection(imagenet_classes:list, things_objects:list) -> set:
     return set(things_objects).intersection(set(imagenet_classes))
 
-def get_cls_mapping_imgnet(PATH:str, filename:str, save_as_json:bool) -> dict:
+def get_cls_mapping_imgnet(PATH:str, save_as_json:bool=False) -> dict:
     """store ImageNet classes in a idx2cls dictionary, and subsequently save as .json file"""
-    if re.search(r'synset', filename):
-        imagenet_classes = parse_imagenet_synsets(filename, PATH)
+    if re.search(r'synset', PATH.split('/')[-1]):
+        imagenet_classes = parse_imagenet_synsets(PATH)
     else:
-        imagenet_classes = parse_imagenet_classes(filename, PATH)
+        imagenet_classes = parse_imagenet_classes(PATH)
     idx2cls = dict(enumerate(imagenet_classes))
     if save_as_json:
         filename = 'imagenet_idx2class.json'
+        PATH = '/'.join(PATH.split('/')[:-1])
         with open(pjoin(PATH, filename), 'w') as f:
             json.dump(idx2cls, f)
     return idx2cls
 
+def get_imagenet_classes(PATH:str, predictions:np.ndarray, save_as_json:bool=False) -> List[str]:
+    idx2cls = get_cls_mapping_imgnet(PATH, save_as_json)
+    return [idx2cls[pred] for pred in predictions]
+
 def json2dict(PATH:str, filename:str) -> dict:
     with open(pjoin(PATH, filename), 'r') as f:
         idx2cls = dict(json.load(f))
@@ -446,9 +502,11 @@ def save_targets(targets:np.ndarray, out_path:str, file_format:str) -> None:
     if re.search(r'npy', file_format):
         with open(pjoin(out_path, 'targets.npy'), 'wb') as f:
             np.save(f, targets)
+    elif re.search(r'mat', file_format):
+        scipy.io.savemat(pjoin(out_path, 'targets.mat'), {'targets': targets})
     else:
         np.savetxt(pjoin(out_path, 'targets.txt'), targets)
-    print(f'\nTargets successfully saved to disk.\n')
+    print(f'...Targets successfully saved to disk.\n')
 
 #############################################################################################################
 ################################ HELPER FUNCTIONS FOR RSA & RDM COMPUTATIONS ################################
@@ -493,7 +551,7 @@ def compute_rdm(F:np.ndarray, method:str='correlation') -> np.ndarray:
     assert method in methods, f'\nMethod to compute RDM must be one of {methods}.\n'
     if method == 'euclidean':
         rdm = euclidean_matrix(F)
-        return rdm / rdm.max()
+        return rdm
     else:
         if method == 'correlation':
             rsm =  correlation_matrix(F)
@@ -504,10 +562,10 @@ def compute_rdm(F:np.ndarray, method:str='correlation') -> np.ndarray:
     return 1 - rsm
 
 def correlate_rdms(rdm_1:np.ndarray, rdm_2:np.ndarray, correlation:str='pearson') -> float:
-    #since RDMs are symmetric matrices, we only need to compare their lower triangular parts (main diagonal can be omitted)
-    tril_inds = np.tril_indices(len(rdm_1), k=-1)
+    #since RDMs are symmetric matrices, we only need to compare their upper (or lower) triangular parts (main diagonal can be omitted)
+    triu_inds = np.triu_indices(len(rdm_1), k=1)
     corr_func = getattr(scipy.stats, ''.join((correlation, 'r')))
-    rho = corr_func(rdm_1[tril_inds], rdm_2[tril_inds])[0]
+    rho = corr_func(rdm_1[triu_inds], rdm_2[triu_inds])[0]
     return rho
 
 def plot_rdm(out_path:str, F:np.ndarray, method:str='correlation', format:str='.png', colormap:str='cividis', show_plot:bool=False) -> None:
@@ -518,7 +576,7 @@ def plot_rdm(out_path:str, F:np.ndarray, method:str='correlation', format:str='.
     plt.yticks([])
     plt.tight_layout()
     if not os.path.exists(out_path):
-        print(f'\nOutput directory did not exists. Creating directories...\n')
+        print(f'\n...Output directory did not exists. Creating directories.\n')
         os.makedirs(out_path)
     plt.savefig(os.path.join(out_path, ''.join(('rdm', format))))
     if show_plot:
@@ -572,17 +630,16 @@ def get_features(
                 out_path:str,
                 model_names:List[str],
                 module_names:List[str],
+                clip:List[bool],
                 pretrained:bool,
                 batch_size:int,
                 flatten_acts:bool,
-                clip:List[bool],
-) -> Dict[str, dict]:
+) -> Dict[str, Dict[str, np.ndarray]]:
     device = 'cuda' if torch.cuda.is_available() else 'cpu'
     model_features = defaultdict(dict)
     for i, model_name in enumerate(model_names):
-        device = device if clip[i] else torch.device(device)
         model, transforms = load_model(model_name, pretrained=pretrained, model_path=None, device=device)
-        dl = load_dl(root, out_path, batch_size=batch_size, transforms=transforms)
+        dl = load_dl(root, out_path=out_path, batch_size=batch_size, transforms=transforms)
         features, _ = extract_features(model, dl, module_names[i], batch_size=batch_size, flatten_acts=flatten_acts, device=device, clip=clip[i])
         model_features[model_name][module_names[i]] = features
     return model_features
@@ -592,10 +649,10 @@ def compare_models_to_humans(
                             out_path:str,
                             model_names:List[str],
                             module_names:List[str],
+                            clip:List[bool],
                             pretrained:bool,
                             batch_size:int,
                             flatten_acts:bool,
-                            clip:List[bool],
                             human_rdm:np.ndarray,
                             save_features:bool=True,
                             n_bootstraps:int=1000,
@@ -608,7 +665,6 @@ def compare_models_to_humans(
     #save model features to disc
     if save_features:
         pickle_file_(model_features, out_path, 'features')
-
     #compare features of each model combination for N bootstraps
     scores = defaultdict(lambda: defaultdict(dict))
     model_combs = list(itertools.combinations(model_names, 2))
@@ -650,10 +706,9 @@ def compare_models(
     #save model features to disc
     if save_features:
         pickle_file_(model_features, out_path, 'features')
-
     #compare features of each model combination for N bootstraps
-    corrs = pd.DataFrame(index=np.arange(len(model_names)), columns=model_names, dtype=float)
-    model_combs = list(itertools.product(model_names, model_names))
+    corrs = pd.DataFrame(np.eye(len(model_names)), index=np.arange(len(model_names)), columns=model_names, dtype=float)
+    model_combs = list(itertools.combinations(model_names, 2))
     for (model_i, model_j) in model_combs:
         module_i = module_names[model_names.index(model_i)]
         module_j = module_names[model_names.index(model_j)]
@@ -663,7 +718,8 @@ def compare_models(
         rdm_j = compute_rdm(features_j, dissimilarity)
         corr = correlate_rdms(rdm_i, rdm_j, correlation)
         corrs.loc[model_names.index(model_i), model_j] = corr
-    corrs['model_names'] = df.columns.to_list()
+        corrs.loc[model_names.index(model_j), model_i] = corr
+    corrs['model_names'] = corrs.columns.to_list()
     corrs.set_index('model_names', inplace=True, drop=True)
     return corrs
 
